@@ -40,18 +40,23 @@ to do better in practice:
 
 class Minimizer(object):
 
-    def __init__(self, initial, condition, random):
+    def __init__(self, initial, condition, random, cautious):
         self.current = hbytes(initial)
         self.size = len(self.current)
         self.condition = condition
         self.random = random
+        self.cautious = cautious
         self.changes = 0
+        self.seen = set()
 
     def incorporate(self, buffer):
         assert isinstance(buffer, hbytes)
         assert len(buffer) == self.size
         assert buffer <= self.current
-        if self.condition(buffer):
+        if buffer in self.seen:
+            return False
+        self.seen.add(buffer)
+        if buffer != self.current and self.condition(buffer):
             self.current = buffer
             self.changes += 1
             return True
@@ -60,15 +65,19 @@ class Minimizer(object):
     def _shrink_index(self, i, c):
         assert isinstance(self.current, hbytes)
         assert 0 <= i < self.size
-        if self.current[i] <= c:
-            return False
+        assert self.current[i] > c
         if self.incorporate(
             self.current[:i] + hbytes([c]) +
             self.current[i + 1:]
         ):
             return True
+
+        if self.cautious:
+            return False
+
         if i == self.size - 1:
             return False
+
         return self.incorporate(
             self.current[:i] + hbytes([c, 255]) +
             self.current[i + 2:]
@@ -77,32 +86,85 @@ class Minimizer(object):
             hbytes([255] * (self.size - i - 1))
         )
 
+    def ddzero(self):
+        self.ddfixate(lambda b: 0)
+
+    def ddshift(self):
+        self.ddfixate(lambda b: b >> 1)
+
+    def ddfixate(self, f):
+        prev = -1
+        while prev != self.changes:
+            prev = self.changes
+            k = len(self.current)
+            while k > 0:
+                i = 0
+                while i + k <= len(self.current):
+                    self.incorporate(
+                        self.current[:i] +
+                        hbytes(f(b) for b in self.current[i:i + k]) +
+                        self.current[i + k:]
+                    )
+                    i += k
+                k //= 2
+
+    def rotate_suffixes(self):
+        for significant, c in enumerate(self.current):  # pragma: no branch
+            if c:
+                break
+        assert self.current[significant]
+
+        prefix = hbytes(significant)
+
+        for i in range(1, self.size - significant):
+            left = self.current[significant:significant + i]
+            right = self.current[significant + i:]
+            rotated = prefix + right + left
+            if rotated < self.current:
+                self.incorporate(rotated)
+
+    def shrink_indices(self):
+        for i in hrange(self.size):
+            t = self.current[i]
+            if t > 0:
+                ss = small_shrinks[self.current[i]]
+                for c in ss:
+                    if self._shrink_index(i, c):
+                        for c in hrange(self.current[i]):
+                            if c in ss:
+                                continue
+                            if self._shrink_index(i, c):
+                                break
+                        break
+
     def run(self):
         if not any(self.current):
             return
+        if len(self.current) == 1:
+            for c in range(self.current[0]):
+                if self.incorporate(hbytes([c])):
+                    break
+            return
         if self.incorporate(hbytes(self.size)):
             return
-        for c in hrange(max(self.current)):
-            if self.incorporate(
-                hbytes(min(b, c) for b in self.current)
-            ):
-                break
-
+        if self.incorporate(hbytes([0] * (self.size - 1) + [1])):
+            return
         change_counter = -1
         while self.current and change_counter < self.changes:
             change_counter = self.changes
-            for i in hrange(self.size):
-                t = self.current[i]
-                if t > 0:
-                    ss = small_shrinks[self.current[i]]
-                    for c in ss:
-                        if self._shrink_index(i, c):
-                            for c in hrange(self.current[i]):
-                                if c in ss:
-                                    continue
-                                if self._shrink_index(i, c):
-                                    break
-                            break
+
+            self.ddzero()
+            self.ddshift()
+
+            if change_counter != self.changes:
+                continue
+
+            self.shrink_indices()
+
+            if change_counter != self.changes or self.cautious:
+                continue
+
+            self.rotate_suffixes()
 
 
 # Table of useful small shrinks to apply to a number.
@@ -123,10 +185,21 @@ for b in hrange(10, 256):
         result.add(b ^ (1 << i))
     result.discard(b)
     assert len(result) <= 10
-    small_shrinks.append(sorted(result))
+    small_shrinks.append(sorted([c for c in result if c < b]))
 
 
-def minimize(initial, condition, random=None):
-    m = Minimizer(initial, condition, random)
+def minimize(initial, condition, random=None, cautious=False):
+    """Perform a lexicographical minimization of the byte string 'initial' such
+    that the predicate 'condition' returns True, and return the minimized
+    string.
+
+    If 'cautious' is set to True, this will only consider strings that
+    satisfy the stronger condition that no individual byte is increased
+    from the original. e.g. normally bytes([0, 255]) is considered a
+    valid shrink of bytes([1, 0]), but if cautious is set it will not
+    be.
+
+    """
+    m = Minimizer(initial, condition, random, cautious)
     m.run()
     return m.current
