@@ -1,12 +1,13 @@
 # coding=utf-8
 #
-# This file is part of Hypothesis (https://github.com/DRMacIver/hypothesis)
+# This file is part of Hypothesis, which may be found at
+# https://github.com/HypothesisWorks/hypothesis-python
 #
-# Most of this work is copyright (C) 2013-2015 David R. MacIver
+# Most of this work is copyright (C) 2013-2016 David R. MacIver
 # (david@drmaciver.com), but it contains contributions by others. See
-# https://github.com/DRMacIver/hypothesis/blob/master/CONTRIBUTING.rst for a
-# full list of people who may hold copyright, and consult the git log if you
-# need to determine who owns an individual contribution.
+# CONTRIBUTING.rst for a full list of people who may hold copyright, and
+# consult the git log if you need to determine who owns an individual
+# contribution.
 #
 # This Source Code Form is subject to the terms of the Mozilla Public License,
 # v. 2.0. If a copy of the MPL was not distributed with this file, You can
@@ -29,8 +30,9 @@ from types import ModuleType
 from functools import wraps
 
 from hypothesis.configuration import storage_directory
-from hypothesis.internal.compat import hrange, qualname, text_type, \
-    getargspec, to_unicode, isidentifier, unicode_safe_repr, \
+from hypothesis.vendor.pretty import pretty
+from hypothesis.internal.compat import hrange, to_str, qualname, \
+    getargspec, to_unicode, isidentifier, str_to_bytes, \
     ARG_NAME_ATTRIBUTE, update_code_location
 
 
@@ -38,7 +40,7 @@ def fully_qualified_name(f):
     """Returns a unique identifier for f pointing to the module it was define
     on, and an containing functions."""
     if f.__module__ is not None:
-        return f.__module__ + u'.' + qualname(f)
+        return f.__module__ + '.' + qualname(f)
     else:
         return qualname(f)
 
@@ -53,20 +55,20 @@ def function_digest(function):
     """
     hasher = hashlib.md5()
     try:
-        hasher.update(to_unicode(inspect.getsource(function)).encode(u'utf-8'))
+        hasher.update(to_unicode(inspect.getsource(function)).encode('utf-8'))
     # Different errors on different versions of python. What fun.
     except (OSError, IOError, TypeError):
         pass
     try:
-        hasher.update(function.__name__.encode(u'utf-8'))
+        hasher.update(str_to_bytes(function.__name__))
     except AttributeError:
         pass
     try:
-        hasher.update(function.__module__.__name__.encode(u'utf-8'))
+        hasher.update(function.__module__.__name__.encode('utf-8'))
     except AttributeError:
         pass
     try:
-        hasher.update(repr(getargspec(function)).encode(u'utf-8'))
+        hasher.update(str_to_bytes(repr(getargspec(function))))
     except TypeError:
         pass
     return hasher.digest()
@@ -104,18 +106,18 @@ def convert_keyword_arguments(function, args, kwargs):
             elif arg_name in defaults:
                 new_args.append(defaults[arg_name])
             else:
-                raise TypeError(u'No value provided for argument %r' % (
+                raise TypeError('No value provided for argument %r' % (
                     arg_name
                 ))
 
     if kwargs and not argspec.keywords:
         if len(kwargs) > 1:
-            raise TypeError(u'%s() got unexpected keyword arguments %s' % (
-                function.__name__, u', '.join(map(repr, kwargs))
+            raise TypeError('%s() got unexpected keyword arguments %s' % (
+                function.__name__, ', '.join(map(repr, kwargs))
             ))
         else:
             bad_kwarg = next(iter(kwargs))
-            raise TypeError(u'%s() got an unexpected keyword argument %r' % (
+            raise TypeError('%s() got an unexpected keyword argument %r' % (
                 function.__name__, bad_kwarg
             ))
     return tuple(new_args), kwargs
@@ -125,8 +127,7 @@ def convert_positional_arguments(function, args, kwargs):
     """Return a tuple (new_args, new_kwargs) where all possible arguments have
     been moved to kwargs.
 
-    new_args will only be non-empty if function has a
-    variadic argument.
+    new_args will only be non-empty if function has a variadic argument.
 
     """
     argspec = getargspec(function)
@@ -135,7 +136,7 @@ def convert_positional_arguments(function, args, kwargs):
         for k in kwargs.keys():
             if k not in argspec.args:
                 raise TypeError(
-                    u'%s() got an unexpected keyword argument %r' % (
+                    '%s() got an unexpected keyword argument %r' % (
                         function.__name__, k
                     ))
     if len(args) < len(argspec.args):
@@ -143,13 +144,13 @@ def convert_positional_arguments(function, args, kwargs):
             len(args), len(argspec.args) - len(argspec.defaults or ())
         ):
             if argspec.args[i] not in kwargs:
-                raise TypeError(u'No value provided for argument %s' % (
+                raise TypeError('No value provided for argument %s' % (
                     argspec.args[i],
                 ))
 
     if len(args) > len(argspec.args) and not argspec.varargs:
         raise TypeError(
-            u'%s() takes at most %d positional arguments (%d given)' % (
+            '%s() takes at most %d positional arguments (%d given)' % (
                 function.__name__, len(argspec.args), len(args)
             )
         )
@@ -157,7 +158,7 @@ def convert_positional_arguments(function, args, kwargs):
     for arg, name in zip(args, argspec.args):
         if name in kwargs:
             raise TypeError(
-                u'%s() got multiple values for keyword argument %r' % (
+                '%s() got multiple values for keyword argument %r' % (
                     function.__name__, name
                 ))
         else:
@@ -208,13 +209,13 @@ def extract_lambda_source(f):
     bad_lambda = False
     for a in args:
         if isinstance(a, (tuple, list)):  # pragma: no cover
-            arg_strings.append(u'(%s)' % (u', '.join(a),))
+            arg_strings.append('(%s)' % (', '.join(a),))
             bad_lambda = True
         else:
             assert isinstance(a, str)
             arg_strings.append(a)
 
-    if_confused = u'lambda %s: <unknown>' % (u', '.join(arg_strings),)
+    if_confused = 'lambda %s: <unknown>' % (', '.join(arg_strings),)
     if bad_lambda:  # pragma: no cover
         return if_confused
     try:
@@ -222,18 +223,16 @@ def extract_lambda_source(f):
     except IOError:
         return if_confused
 
-    if not isinstance(source, text_type):  # pragma: no branch
-        source = source.decode(u'utf-8')  # pragma: no cover
-    source = LINE_CONTINUATION.sub(u' ', source)
-    source = WHITESPACE.sub(u' ', source)
+    source = LINE_CONTINUATION.sub(' ', source)
+    source = WHITESPACE.sub(' ', source)
     source = source.strip()
 
     try:
         tree = ast.parse(source)
     except SyntaxError:
-        for i in hrange(len(source) - 1, len(u'lambda'), -1):
+        for i in hrange(len(source) - 1, len('lambda'), -1):
             prefix = source[:i]
-            if u'lambda' not in prefix:
+            if 'lambda' not in prefix:
                 return if_confused
             try:
                 tree = ast.parse(prefix)
@@ -255,8 +254,8 @@ def extract_lambda_source(f):
     assert lambda_ast.lineno == 1
     source = source[lambda_ast.col_offset:].strip()
 
-    source = source[source.index(u'lambda'):]
-    for i in hrange(len(source), len(u'lambda'), -1):  # pragma: no branch
+    source = source[source.index('lambda'):]
+    for i in hrange(len(source), len('lambda'), -1):  # pragma: no branch
         try:
             parsed = ast.parse(source[:i])
             assert len(parsed.body) == 1
@@ -267,28 +266,28 @@ def extract_lambda_source(f):
             break
         except SyntaxError:
             pass
-    lines = source.split(u'\n')
-    lines = [PROBABLY_A_COMMENT.sub(u'', l) for l in lines]
-    source = u'\n'.join(lines)
+    lines = source.split('\n')
+    lines = [PROBABLY_A_COMMENT.sub('', l) for l in lines]
+    source = '\n'.join(lines)
 
-    source = WHITESPACE.sub(u' ', source)
-    source = SPACE_FOLLOWS_OPEN_BRACKET.sub(u'(', source)
-    source = SPACE_PRECEDES_CLOSE_BRACKET.sub(u')', source)
+    source = WHITESPACE.sub(' ', source)
+    source = SPACE_FOLLOWS_OPEN_BRACKET.sub('(', source)
+    source = SPACE_PRECEDES_CLOSE_BRACKET.sub(')', source)
     source = source.strip()
     return source
 
 
 def get_pretty_function_description(f):
-    if not hasattr(f, u'__name__'):
+    if not hasattr(f, '__name__'):
         return repr(f)
     name = f.__name__
-    if name == u'<lambda>':
+    if name == '<lambda>':
         result = extract_lambda_source(f)
         return result
     elif isinstance(f, types.MethodType):
         self = f.__self__
         if not (self is None or inspect.isclass(self)):
-            return u'%r.%s' % (self, name)
+            return '%r.%s' % (self, name)
     return name
 
 
@@ -298,7 +297,7 @@ def nicerepr(v):
     elif isinstance(v, type):
         return v.__name__
     else:
-        return unicode_safe_repr(v)
+        return to_str(pretty(v))
 
 
 def arg_string(f, args, kwargs, reorder=True):
@@ -311,12 +310,12 @@ def arg_string(f, args, kwargs, reorder=True):
 
     for a in argspec.args:
         if a in kwargs:
-            bits.append(u'%s=%s' % (a, nicerepr(kwargs.pop(a))))
+            bits.append('%s=%s' % (a, nicerepr(kwargs.pop(a))))
     if kwargs:
         for a in sorted(kwargs):
-            bits.append(u'%s=%s' % (a, nicerepr(kwargs[a])))
+            bits.append('%s=%s' % (a, nicerepr(kwargs[a])))
 
-    return u', '.join(
+    return ', '.join(
         [nicerepr(x) for x in args] +
         bits
     )
@@ -325,12 +324,12 @@ def arg_string(f, args, kwargs, reorder=True):
 def unbind_method(f):
     """Take something that might be a method or a function and return the
     underlying function."""
-    return getattr(f, u'im_func', getattr(f, u'__func__', f))
+    return getattr(f, 'im_func', getattr(f, '__func__', f))
 
 
 def check_valid_identifier(identifier):
     if not isidentifier(identifier):
-        raise ValueError(u'%r is not a valid python identifier' %
+        raise ValueError('%r is not a valid python identifier' %
                          (identifier,))
 
 
@@ -348,8 +347,9 @@ def source_exec_as_module(source):
         pass
 
     result = ModuleType('hypothesis_temporary_module_%s' % (
-        hashlib.sha1(source.encode(u'utf-8')).hexdigest(),
+        hashlib.sha1(str_to_bytes(source)).hexdigest(),
     ))
+    assert isinstance(source, str)
     exec(source, result.__dict__)
     eval_cache[source] = result
     return result
@@ -362,12 +362,12 @@ def accept(%(funcname)s):
     def %(name)s(%(argspec)s):
         return %(funcname)s(%(invocation)s)
     return %(name)s
-""".strip() + u'\n'
+""".strip() + '\n'
 
 
-def copy_argspec(name, argspec):
-    """A decorator which sets the name and argspec of the function passed into
-    it."""
+def define_function_signature(name, docstring, argspec):
+    """A decorator which sets the name, argspec and docstring of the function
+    passed into it."""
     check_valid_identifier(name)
     for a in argspec.args:
         check_valid_identifier(a)
@@ -381,7 +381,7 @@ def copy_argspec(name, argspec):
         for a in argspec.args[:-n_defaults]:
             parts.append(a)
         for a in argspec.args[-n_defaults:]:
-            parts.append(u'%s=not_set' % (a,))
+            parts.append('%s=not_set' % (a,))
     else:
         parts = list(argspec.args)
     used_names = list(argspec.args)
@@ -398,18 +398,18 @@ def copy_argspec(name, argspec):
                 invocation_parts.append(a)
         if argspec.varargs:
             used_names.append(argspec.varargs)
-            parts.append(u'*' + argspec.varargs)
-            invocation_parts.append(u'*' + argspec.varargs)
+            parts.append('*' + argspec.varargs)
+            invocation_parts.append('*' + argspec.varargs)
         for k in must_pass_as_kwargs:
-            invocation_parts.append(u'%(k)s=%(k)s' % {u'k': k})
+            invocation_parts.append('%(k)s=%(k)s' % {'k': k})
 
         if argspec.keywords:
             used_names.append(argspec.keywords)
-            parts.append(u'**' + argspec.keywords)
-            invocation_parts.append(u'**' + argspec.keywords)
+            parts.append('**' + argspec.keywords)
+            invocation_parts.append('**' + argspec.keywords)
 
-        candidate_names = [u'f'] + [
-            u'f_%d' % (i,) for i in hrange(1, len(used_names) + 2)
+        candidate_names = ['f'] + [
+            'f_%d' % (i,) for i in hrange(1, len(used_names) + 2)
         ]
 
         for funcname in candidate_names:  # pragma: no branch
@@ -418,13 +418,14 @@ def copy_argspec(name, argspec):
 
         base_accept = source_exec_as_module(
             COPY_ARGSPEC_SCRIPT % {
-                u'name': name,
-                u'funcname': funcname,
-                u'argspec': u', '.join(parts),
-                u'invocation': u', '.join(invocation_parts)
+                'name': name,
+                'funcname': funcname,
+                'argspec': ', '.join(parts),
+                'invocation': ', '.join(invocation_parts)
             }).accept
 
         result = base_accept(f)
+        result.__doc__ = docstring
         result.__defaults__ = argspec.defaults
         return result
     return accept
@@ -453,5 +454,6 @@ def impersonate(target):
 def proxies(target):
     def accept(proxy):
         return impersonate(target)(wraps(target)(
-            copy_argspec(target.__name__, getargspec(target))(proxy)))
+            define_function_signature(
+                target.__name__, target.__doc__, getargspec(target))(proxy)))
     return accept
