@@ -25,8 +25,8 @@ from fractions import Fraction
 from hypothesis.errors import InvalidArgument
 from hypothesis.control import assume
 from hypothesis.searchstrategy import SearchStrategy
-from hypothesis.internal.compat import ArgSpec, text_type, getargspec, \
-    integer_types, float_to_decimal
+from hypothesis.internal.compat import hrange, ArgSpec, text_type, \
+    getargspec, integer_types, float_to_decimal
 from hypothesis.internal.floats import is_negative, float_to_int, \
     int_to_float, count_between_floats
 from hypothesis.utils.conventions import not_set
@@ -128,6 +128,7 @@ class Nothing(SearchStrategy):
     def flatmap(self, f):
         return self
 
+
 NOTHING = Nothing()
 
 
@@ -173,16 +174,18 @@ def one_of(*args):
         except TypeError:
             pass
 
+    strategies = []
     for arg in args:
         check_strategy(arg)
-    args = [a for a in args if not a.is_empty]
+        if not arg.is_empty:
+            strategies.extend([s for s in arg.branches if not s.is_empty])
 
-    if not args:
+    if not strategies:
         return nothing()
-    if len(args) == 1:
-        return args[0]
+    if len(strategies) == 1:
+        return strategies[0]
     from hypothesis.searchstrategy.strategies import OneOfStrategy
-    return OneOfStrategy(args)
+    return OneOfStrategy(strategies)
 
 
 @cacheable
@@ -800,21 +803,25 @@ def fractions(min_value=None, max_value=None, max_denominator=None):
 
 @cacheable
 @defines_strategy
-def decimals(min_value=None, max_value=None):
+def decimals(
+    min_value=None, max_value=None, allow_nan=None, allow_infinity=None
+):
     """Generates instances of decimals.Decimal.
 
-    If min_value is not None then all generated values are no less than
-    min_value.
+    - If min_value is not None, all values will be >= min_value.
+    - If max_value is not None, all values will be <= max_value.
+    - If min_value or max_value is not None, it is an error to enable
+      allow_nan.
+    - If both min_value and max_value are not None, it is an error to enable
+      allow_infinity.
 
-    If max_value is not None then all generated values are no greater than
-    max_value.
+    Where not explicitly ruled out by the bounds, all of infinity, -infinity
+    and NaN are possible values generated by this strategy.
 
     """
-    check_valid_bound(min_value, 'min_value')
-    check_valid_bound(max_value, 'max_value')
-    check_valid_interval(min_value, max_value, 'min_value', 'max_value')
-
-    floats_strategy = floats(min_value=min_value, max_value=max_value)
+    floats_strategy = floats(
+        min_value=min_value, max_value=max_value,
+        allow_nan=allow_nan, allow_infinity=allow_infinity)
     fractions_strategy = fractions(min_value=min_value, max_value=max_value)
 
     return (
@@ -850,18 +857,24 @@ def recursive(base, extend, max_leaves=100):
 def permutations(values):
     """Return a strategy which returns permutations of the collection
     "values"."""
+    from hypothesis.internal.conjecture.utils import integer_range
+
     values = list(values)
     if not values:
-        return just(()).map(lambda _: [])
+        return builds(list)
 
-    def build_permutation(swaps):
-        initial = list(values)
-        for i, j in swaps:
-            initial[i], initial[j] = initial[j], initial[i]
-        return initial
-    n = len(values)
-    index = integers(0, n - 1)
-    return lists(tuples(index, index), max_size=n ** 2).map(build_permutation)
+    class PermutationStrategy(SearchStrategy):
+
+        def do_draw(self, data):
+            # Reversed Fisher-Yates shuffle. Reverse order so that it shrinks
+            # propertly: This way we prefer things that are lexicographically
+            # closer to the identity.
+            result = list(values)
+            for i in hrange(len(result)):
+                j = integer_range(data, i, len(result) - 1)
+                result[i], result[j] = result[j], result[i]
+            return result
+    return PermutationStrategy()
 
 
 @cacheable
@@ -874,7 +887,7 @@ def composite(f):
 
     """
 
-    from hypothesis.internal.reflection import copy_argspec
+    from hypothesis.internal.reflection import define_function_signature
     argspec = getargspec(f)
 
     if (
@@ -895,12 +908,20 @@ def composite(f):
     )
 
     @defines_strategy
-    @copy_argspec(f.__name__, new_argspec)
+    @define_function_signature(f.__name__, f.__doc__, new_argspec)
     def accept(*args, **kwargs):
         class CompositeStrategy(SearchStrategy):
 
             def do_draw(self, data):
-                return f(data.draw, *args, **kwargs)
+                first_draw = [True]
+
+                def draw(strategy):
+                    if not first_draw[0]:
+                        data.mark_bind()
+                    first_draw[0] = False
+                    return data.draw(strategy)
+
+                return f(draw, *args, **kwargs)
         return CompositeStrategy()
     return accept
 
@@ -1030,10 +1051,14 @@ def data():
         def __repr__(self):
             return 'data(...)'
 
-        def draw(self, strategy):
+        def draw(self, strategy, label=None):
+            self.data.mark_bind()
             result = self.data.draw(strategy)
             self.count += 1
-            note('Draw %d: %r' % (self.count, result))
+            if label is not None:
+                note('Draw %d (%s): %r' % (self.count, label, result))
+            else:
+                note('Draw %d: %r' % (self.count, result))
             return result
 
     class DataStrategy(SearchStrategy):
